"""
In this script, a one dimensional MSO (multiple superimposed sine wave) signal
is created.
"""


import numpy as np
import os

##############
# PARAMETERS #
##############

train_sequence_length = 400 + 1  # [:-1] as input and [1:] as target
test_sequence_length = 400 + 1
test_long_sequence_length = 2000 + 1
frequencies = [0.2, 0.311, 0.42, 0.51, 0.63, 0.74,
               0.85, 0.97, 1.08, 1.19, 1.27, 1.32]
train_samples = 10000
test_samples = 1000
test_long_samples = 100
number_of_waves = 5

# Define seed for random number generator
np.random.seed(1234)

#############
# FUNCTIONS #
#############


def generate_random_sample(number_of_waves, freqs, sequence_length):
    """
    This function generates random amplitudes and phase shifts and returns a
    corresponding sample
    """

    # Randomly create amplitudes in [0, 1) and phases in [0, 2pi) for each wave
    # signal of the MSO
    amps = np.random.rand(number_of_waves)
    phases = np.random.rand(number_of_waves) * 2*np.pi

    # Generate and return a sample with the just created amplitudes and phases
    return generate_sample(number_of_waves=number_of_waves, freqs=freqs,
                           amps=amps, phases=phases,
                           sequence_length=sequence_length)


def generate_sample(number_of_waves, freqs, amps, phases, sequence_length):
    """
    This function receives particular frequencies, amplitudes and phase shifts
    to produce and return a MSO signal (number of superimposed oscillators is
    determined by the number_of_waves argument).
    """

    # Initialize empty array to store the MSO signal sample
    ret = np.zeros(sequence_length, dtype=np.float32)

    # Iterate over the entire sequence length
    for t in range(sequence_length):

        # Initialize a variable to store the current time step's wave signal for
        # each MSO
        mso_signal = 0

        # Iterate over all MSO signals and add them to the overall mso_signal
        for _i in range(number_of_waves):
            mso_signal += amps[_i] * np.sin(freqs[_i] * t + phases[_i])

        ret[t] = mso_signal

    return ret


def generate_samples(number_of_samples, number_of_waves, frequencies,
                     sequence_length, add_canonical_example=False):

    samples = []

    for n in range(number_of_samples):

        if n == 0 and add_canonical_example:
            # Let the first sample always be a default sample with constant
            # amplitudes (ones) and without phase shifts
            amplitudes = np.ones(number_of_waves)
            phase_shifts = np.zeros(number_of_waves)

            # Generate the default sample
            sample = generate_sample(number_of_waves=number_of_waves,
                                     freqs=frequencies,
                                     amps=amplitudes,
                                     phases=phase_shifts,
                                     sequence_length=sequence_length)
        else:
            # Create a sample with random amplitudes and phase shifts
            sample = generate_random_sample(number_of_waves=number_of_waves,
                                            freqs=frequencies,
                                            sequence_length=sequence_length)

        samples.append(sample)

    return samples


##########
# SCRIPT #
##########

# Set up paths to store the data
path_train = "data/train/"
path_test = "data/test/"
path_test_long = "data/test_long/"

# Statement to console
print("Generating data for " + str(number_of_waves) + " wave(s)")

# Create the data
train_data = generate_samples(number_of_samples=train_samples,
                              number_of_waves=number_of_waves,
                              frequencies=frequencies,
                              sequence_length=train_sequence_length,
                              add_canonical_example=True)

test_data = generate_samples(number_of_samples=test_samples,
                             number_of_waves=number_of_waves,
                             frequencies=frequencies,
                             sequence_length=test_sequence_length)

test_long_data = generate_samples(number_of_samples=test_long_samples,
                                  number_of_waves=number_of_waves,
                                  frequencies=frequencies,
                                  sequence_length=test_long_sequence_length,
                                  add_canonical_example=True)

# Create appropriate train and test file names including the MSO-complexity
file_name_train = "mso-" + str(number_of_waves) + "_train"
file_name_test = "mso-" + str(number_of_waves) + "_test"
file_name_test_long = "mso-" + str(number_of_waves) + "_test_long"

# Create target directories of yet existing
os.makedirs(path_train, exist_ok=True)
os.makedirs(path_test, exist_ok=True)
os.makedirs(path_test_long, exist_ok=True)

# Write the samples containing the current MSO-complexity to file
np.save(path_train + file_name_train, np.array(train_data))
np.save(path_test + file_name_test, np.array(test_data))
np.save(path_test_long + file_name_test_long, np.array(test_long_data))
