"""
In this script, a one dimensional MSO (multiple superimposed sine wave) signal
is created.
"""


import numpy as np
import os
import pendulum

##############
# PARAMETERS #
##############

pendulum_trajectory_length = 4000

train_sequence_length = 400 + 1
test_sequence_length = 400 + 1
test_long_sequence_length = 2000 + 1

train_samples = 10000
test_samples = 1000
test_long_samples = 100

pendulum_l1 = 1.0
pendulum_l2 = 1.0
pendulum_m1 = 1.0
pendulum_m2 = 1.0

pendulum_theta1_min = np.pi * 0.5
pendulum_theta1_max = np.pi * 0.5 + np.pi
pendulum_theta2_rel_angle_max = np.deg2rad(30)
pendulum_delta_t = 0.01

# Define seed for random number generator
np.random.seed(1234)

#############
# FUNCTIONS #
#############


def generate_trajectory(pendulum, trajectory_length):
    """
    """

    ret = np.zeros((trajectory_length, 2), dtype=np.float32)
    
    for t in range(trajectory_length):
        ret[t][0] = pendulum.x2
        ret[t][1] = pendulum.y2
        pendulum.move(pendulum_delta_t)

    return ret
    

def generate_random_trajectory(trajectory_length):
    """

    """
    theta1 = pendulum_theta1_min + np.random.rand() * (
        pendulum_theta1_max - pendulum_theta1_min
    )
    theta2 = theta1 + (np.random.rand() - 0.5) \
             * 2.0 * pendulum_theta2_rel_angle_max

    pen = pendulum.Pendulum(
        l1=pendulum_l1, l2=pendulum_l2, 
        m1=pendulum_m1, m2=pendulum_m2,
        theta1=theta1, theta2=theta2,
        theta1dot=0.0, theta2dot=0.0)
    
    return generate_trajectory(pendulum=pen,
                               trajectory_length=trajectory_length)


def generate_samples(number_of_samples, trajectory_length, sequence_length):

    # Initialize a zero array to store all samples
    samples = np.zeros((number_of_samples, sequence_length, 2))  # 100, 2001

    # Determine the samples that are generated from one pendulum simulation and
    # the trajectory length of one sample
    s_per_traj = int(np.ceil(trajectory_length / sequence_length))  # 2
    trajectory_length += s_per_traj  # 4002

    for i in range(number_of_samples // s_per_traj):  # 0, 1, ..., 50

        if i % 20 == 0:
            print("Generating sample " + str(i * s_per_traj) + "/"
                  + str(number_of_samples))

        # Create a trajectory that can be splitted into s_per_traj many samples
        sample = generate_random_trajectory(
            trajectory_length=trajectory_length
        )
        sample = np.reshape(sample, (s_per_traj, sequence_length, 2))

        # Insert the sample ito the samples list
        samples[s_per_traj * i : s_per_traj * i + s_per_traj] = sample

    return samples


##########
# SCRIPT #
##########

# Set up paths to store the data
path_train = "data/train/"
path_test = "data/test/"
path_test_long = "data/test_long/"

# Create the data
print("Generation of train_data...")
train_data = generate_samples(number_of_samples=train_samples,
                              trajectory_length=pendulum_trajectory_length,
                              sequence_length=train_sequence_length)

print("Generation of test_data...")
test_data = generate_samples(number_of_samples=test_samples,
                             trajectory_length=pendulum_trajectory_length,
                             sequence_length=test_sequence_length)

print("Generation of test_long_data...")
test_long_data = generate_samples(number_of_samples=test_long_samples,
                                  trajectory_length=pendulum_trajectory_length,
                                  sequence_length=test_long_sequence_length)

# Create appropriate train and test file names including the MSO-complexity
file_name_train = "pendulum_train"
file_name_test = "pendulum_test"
file_name_test_long = "pendulum_test_long"

# Create target directories of yet existing
os.makedirs(path_train, exist_ok=True)
os.makedirs(path_test, exist_ok=True)
os.makedirs(path_test_long, exist_ok=True)

# Write the samples containing the current MSO-complexity to file
np.save(path_train + file_name_train, train_data)
np.save(path_test + file_name_test, test_data)
np.save(path_test_long + file_name_test_long, test_long_data)

