import numpy as np
import os


class Pendulum():
    """


    """

    def __init__(self, l1=1.0, l2=1.0, m1=1.0, m2=1.0, 
                 theta1=np.pi / 2.0, theta2=np.pi / 2.0, 
                 theta1dot=0.0, theta2dot=0.0):

        self.set_parameters(l1=l1, l2=l2, m1=m1, m2=m2,
                            theta1=theta1, theta2=theta2,
                            theta1dot=theta1dot, theta2dot=theta2dot)

        self.g = 9.81

    def set_parameters(self, 
                       l1, l2, m1, m2,
                       theta1, theta2, 
                       theta1dot, theta2dot):
        self.l1 = l1
        self.l2 = l2
        self.m1 = m1
        self.m2 = m2
        self.state = np.array([theta1, theta2, theta1dot, theta2dot])

        self.calculate_positions()
    
    def calculate_positions(self):
        theta1, theta2, _, _ = self.state

        self.x1 = self.l1 * np.sin(theta1)
        self.x2 = self.x1 + self.l2 * np.sin(theta2)
        self.y1 = self.l1 * -np.cos(theta1)
        self.y2 = self.y1 + self.l2 * -np.cos(theta2)

    def derivatives(self, y_n):

        theta1, theta2, omega1, omega2 = y_n
        cos = np.cos(theta2 - theta1)
        sin = np.sin(theta2 - theta1)

        theta1dot = omega1
        theta2dot = omega2

        la = self.l1 / self.l2;
        g1 = self.g / self.l1;
        g2 = self.g / self.l2;
        mu = self.m2 / (self.m1 + self.m2);
        
        omega1dot = (
            mu * g1 * np.sin(theta2) * cos + 
            mu * (theta1dot ** 2) * sin * cos - 
            g1 * np.sin(theta1) + 
            (mu / la) * (theta2dot ** 2) * sin
        ) / (1 - mu * (cos ** 2))

        omega2dot = (
            g2 * np.sin(theta1) * cos -
            mu * (theta2dot ** 2) * sin * cos -
            g2 * np.sin(theta2) -
            la * (theta1dot ** 2) * sin
        ) / (1 - mu * (cos ** 2))

        return np.array([omega1, omega2, omega1dot, omega2dot])


    def move(self, h=0.01):
        #
        # Fourth order Runge-Kutta numerical integration.
        #
        y_n = self.state
        #
        # K1 = h * dydx(x0, y) 
        # 
        K1 = h * self.derivatives(y_n = y_n)
        #
        # K2 = dt * dydx(x0 + 0.5 * dt, y + 0.5 * K1) 
        #
        K2 = h * self.derivatives(y_n = y_n + 0.5 * K1)
        # 
        # K3 = dt * dydx(x0 + 0.5 * dt, y + 0.5 * K2) 
        #
        K3 = h * self.derivatives(y_n = y_n + 0.5 * K2)
        #
        # K4 = dt * dydx(x0 + dt, y + K3) 
        #
        K4 = h * self.derivatives(y_n = y_n + K3)
        #
        # y_{n+1} = y_n + 1/6*(K1 + 2*K2 + 2*K3 + K4)
        #
        y_np = y_n + (1.0 / 6.0) * (K1 + 2 * K2 + 2 * K3 + K4)
        #
        self.state = y_np
        
        self.calculate_positions()


""" TESTING
p = Pendulum(l1=1.0, l2=1.0, m1=1.0, m2=1.0, theta1=np.pi*0.5, theta2=np.pi*0.5, theta1dot=0, theta2dot=0);
p.move(0.01)
print(p.x2, p.y2)

p.move(0.01)
print(p.x2, p.y2)

p.move(0.01)
print(p.x2, p.y2)

p.move(0.01)
print(p.x2, p.y2)
"""
